import json

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class NationBuilderOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.nationbuilder.NationBuilderOAuth2"
    user_data_url = "https://foobar.nationbuilder.com/api/v1/people/me"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "created_at": 1422937981,
            "expires_in": 2592000,
        }
    )
    user_data_body = json.dumps(
        {
            "person": {
                "twitter_followers_count": None,
                "last_name": "Bar",
                "rule_violations_count": 0,
                "linkedin_id": None,
                "recruiter_id": None,
                "membership_expires_at": None,
                "donations_raised_count": 0,
                "last_contacted_at": None,
                "prefix": None,
                "profile_content_html": None,
                "email4": None,
                "email2": None,
                "availability": None,
                "occupation": None,
                "user_submitted_address": None,
                "could_vote_status": None,
                "state_upper_district": None,
                "salesforce_id": None,
                "van_id": None,
                "phone_time": None,
                "profile_content": None,
                "auto_import_id": None,
                "parent_id": None,
                "email4_is_bad": False,
                "twitter_updated_at": None,
                "email3_is_bad": False,
                "bio": None,
                "party_member": None,
                "unsubscribed_at": None,
                "fax_number": None,
                "last_contacted_by": None,
                "active_customer_expires_at": None,
                "federal_donotcall": False,
                "warnings_count": 0,
                "first_supporter_at": "2015-02-02T19:30:28-08:00",
                "previous_party": None,
                "donations_raised_amount_this_cycle_in_cents": 0,
                "call_status_name": None,
                "marital_status": None,
                "facebook_updated_at": None,
                "donations_count": 0,
                "note_updated_at": None,
                "closed_invoices_count": None,
                "profile_headline": None,
                "fire_district": None,
                "mobile_normalized": None,
                "import_id": None,
                "last_call_id": None,
                "donations_raised_amount_in_cents": 0,
                "facebook_address": None,
                "is_profile_private": False,
                "last_rule_violation_at": None,
                "sex": None,
                "full_name": "Foo Bar",
                "last_donated_at": None,
                "donations_pledged_amount_in_cents": 0,
                "primary_email_id": 1,
                "media_market_name": None,
                "capital_amount_in_cents": 500,
                "datatrust_id": None,
                "precinct_code": None,
                "email3": None,
                "religion": None,
                "first_prospect_at": None,
                "judicial_district": None,
                "donations_count_this_cycle": 0,
                "work_address": None,
                "is_twitter_follower": False,
                "email1": "foobar@gmail.com",
                "email": "foobar@gmail.com",
                "contact_status_name": None,
                "mobile_opt_in": True,
                "twitter_description": None,
                "parent": None,
                "tags": [],
                "first_volunteer_at": None,
                "inferred_support_level": None,
                "banned_at": None,
                "first_invoice_at": None,
                "donations_raised_count_this_cycle": 0,
                "is_donor": False,
                "twitter_location": None,
                "email1_is_bad": False,
                "legal_name": None,
                "language": None,
                "registered_at": None,
                "call_status_id": None,
                "last_invoice_at": None,
                "school_sub_district": None,
                "village_district": None,
                "twitter_name": None,
                "membership_started_at": None,
                "subnations": [],
                "meetup_address": None,
                "author_id": None,
                "registered_address": None,
                "external_id": None,
                "twitter_login": None,
                "inferred_party": None,
                "spent_capital_amount_in_cents": 0,
                "suffix": None,
                "mailing_address": None,
                "is_leaderboardable": True,
                "twitter_website": None,
                "nbec_guid": None,
                "city_district": None,
                "church": None,
                "is_profile_searchable": True,
                "employer": None,
                "is_fundraiser": False,
                "email_opt_in": True,
                "recruits_count": 0,
                "email2_is_bad": False,
                "county_district": None,
                "recruiter": None,
                "twitter_friends_count": None,
                "facebook_username": None,
                "active_customer_started_at": None,
                "pf_strat_id": None,
                "locale": None,
                "twitter_address": None,
                "is_supporter": True,
                "do_not_call": False,
                "profile_image_url_ssl": "https://d3n8a8pro7vhmx.cloudfront.net"
                "/assets/icons/buddy.png",
                "invoices_amount_in_cents": None,
                "username": None,
                "donations_amount_in_cents": 0,
                "is_volunteer": False,
                "civicrm_id": None,
                "supranational_district": None,
                "precinct_name": None,
                "invoice_payments_amount_in_cents": None,
                "work_phone_number": None,
                "phone": "213.394.4623",
                "received_capital_amount_in_cents": 500,
                "primary_address": None,
                "is_possible_duplicate": False,
                "invoice_payments_referred_amount_in_cents": None,
                "donations_amount_this_cycle_in_cents": 0,
                "priority_level": None,
                "first_fundraised_at": None,
                "phone_normalized": "2133944623",
                "rnc_regid": None,
                "twitter_id": None,
                "birthdate": None,
                "mobile": None,
                "federal_district": None,
                "donations_to_raise_amount_in_cents": 0,
                "support_probability_score": None,
                "invoices_count": None,
                "nbec_precinct_code": None,
                "website": None,
                "closed_invoices_amount_in_cents": None,
                "home_address": None,
                "school_district": None,
                "support_level": None,
                "demo": None,
                "children_count": 0,
                "updated_at": "2015-02-02T19:30:28-08:00",
                "membership_level_name": None,
                "billing_address": None,
                "is_ignore_donation_limits": False,
                "signup_type": 0,
                "precinct_id": None,
                "rnc_id": None,
                "id": 2,
                "ethnicity": None,
                "is_survey_question_private": False,
                "middle_name": None,
                "author": None,
                "last_fundraised_at": None,
                "state_file_id": None,
                "note": None,
                "submitted_address": None,
                "support_level_changed_at": None,
                "party": None,
                "contact_status_id": None,
                "outstanding_invoices_amount_in_cents": None,
                "page_slug": None,
                "outstanding_invoices_count": None,
                "first_recruited_at": None,
                "county_file_id": None,
                "first_name": "Foo",
                "facebook_profile_url": None,
                "city_sub_district": None,
                "has_facebook": False,
                "is_deceased": False,
                "labour_region": None,
                "state_lower_district": None,
                "dw_id": None,
                "created_at": "2015-02-02T19:30:28-08:00",
                "is_prospect": False,
                "priority_level_changed_at": None,
                "is_mobile_bad": False,
                "overdue_invoices_count": None,
                "ngp_id": None,
                "do_not_contact": False,
                "first_donated_at": None,
                "turnout_probability_score": None,
            },
            "precinct": None,
        }
    )

    def test_login(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_NATIONBUILDER_SLUG": "foobar"})
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.strategy.set_settings({"SOCIAL_AUTH_NATIONBUILDER_SLUG": "foobar"})
        self.do_partial_pipeline()



"""
Soundcloud OAuth2 backend, docs at:
    https://python-social-auth.readthedocs.io/en/latest/backends/soundcloud.html
"""

from typing import Any

from social_core.utils import url_add_parameters

from .oauth import BaseOAuth2


class SoundcloudOAuth2(BaseOAuth2):
    """Soundcloud OAuth authentication backend"""

    name = "soundcloud"
    AUTHORIZATION_URL = "https://soundcloud.com/connect"
    ACCESS_TOKEN_URL = "https://api.soundcloud.com/oauth2/token"
    SCOPE_SEPARATOR = ","
    REDIRECT_STATE = False
    EXTRA_DATA = [
        ("id", "id"),
        ("refresh_token", "refresh_token"),
        ("expires", "expires"),
    ]

    def get_user_details(self, response):
        """Return user details from Soundcloud account"""
        fullname, first_name, last_name = self.get_user_names(response.get("full_name"))
        return {
            "username": response.get("username"),
            "email": response.get("email") or "",
            "fullname": fullname,
            "first_name": first_name,
            "last_name": last_name,
        }

    def user_data(self, access_token: str, *args, **kwargs) -> dict[str, Any] | None:
        """Loads user data from service"""
        return self.get_json(
            "https://api.soundcloud.com/me",
            headers={"Authorization": f"OAuth {access_token}"},
            params={"format": "json"},
        )

    def auth_url(self):
        """Return redirect url"""
        state = None
        if self.STATE_PARAMETER or self.REDIRECT_STATE:
            # Store state in session for further request validation. The state
            # value is passed as state parameter (as specified in OAuth2 spec),
            # but also added to redirect_uri, that way we can still verify the
            # request if the provider doesn't implement the state parameter.
            # Reuse token if any.
            name = f"{self.name}_state"
            state = self.strategy.session_get(name) or self.state_token()
            self.strategy.session_set(name, state)

        params = self.auth_params(state)
        params.update(self.get_scope_argument())
        params.update(self.auth_extra_arguments())
        return url_add_parameters(self.authorization_url(), params, True)



"""
Copyright (c) 2015 Microsoft Open Technologies, Inc.

All rights reserved.

MIT License

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import json
import os
import tempfile
from typing import cast
from unittest import TestCase
from unittest.mock import patch
from urllib.parse import parse_qs

import responses

from social_core.exceptions import AuthMissingParameter

from .oauth import BaseAuthUrlTestMixin, OAuth2Test


class AzureADOAuth2Test(OAuth2Test, BaseAuthUrlTestMixin):
    backend_path = "social_core.backends.azuread.AzureADOAuth2"
    user_data_url = "https://graph.windows.net/me"
    expected_username = "foobar"
    access_token_body = json.dumps(
        {
            "access_token": "foobar",
            "token_type": "bearer",
            "id_token": "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJodHRwczovL"
            "3N0cy53aW5kb3dzLm5ldC83Mjc0MDZhYy03MDY4LTQ4ZmEtOTJiOS1jMmQ"
            "2NzIxMWJjNTAvIiwiaWF0IjpudWxsLCJleHAiOm51bGwsImF1ZCI6IjAyO"
            "WNjMDEwLWJiNzQtNGQyYi1hMDQwLWY5Y2VkM2ZkMmM3NiIsInN1YiI6In"
            "FVOHhrczltSHFuVjZRMzR6aDdTQVpvY2loOUV6cnJJOW1wVlhPSWJWQTg"
            "iLCJ2ZXIiOiIxLjAiLCJ0aWQiOiI3Mjc0MDZhYy03MDY4LTQ4ZmEtOTJi"
            "OS1jMmQ2NzIxMWJjNTAiLCJvaWQiOiI3ZjhlMTk2OS04YjgxLTQzOGMtO"
            "GQ0ZS1hZDZmNTYyYjI4YmIiLCJ1cG4iOiJmb29iYXJAdGVzdC5vbm1pY3"
            "Jvc29mdC5jb20iLCJnaXZlbl9uYW1lIjoiZm9vIiwiZmFtaWx5X25hbWU"
            "iOiJiYXIiLCJuYW1lIjoiZm9vIGJhciIsInVuaXF1ZV9uYW1lIjoiZm9v"
            "YmFyQHRlc3Qub25taWNyb3NvZnQuY29tIiwicHdkX2V4cCI6IjQ3MzMwO"
            "TY4IiwicHdkX3VybCI6Imh0dHBzOi8vcG9ydGFsLm1pY3Jvc29mdG9ubG"
            "luZS5jb20vQ2hhbmdlUGFzc3dvcmQuYXNweCJ9.3V50dHXTZOHj9UWtkn"
            "2g7BjX5JxNe8skYlK4PdhiLz4",
            "expires_in": 3600,
            "expires_on": 1423650396,
            "not_before": 1423646496,
        }
    )
    refresh_token_body = json.dumps(
        {
            "access_token": "foobar-new-token",
            "token_type": "bearer",
            "expires_in": 3600,
            "refresh_token": "foobar-new-refresh-token",
            "scope": "identity",
        }
    )

    def test_login(self) -> None:
        self.do_login()

    def test_partial_pipeline(self) -> None:
        self.do_partial_pipeline()

    def test_refresh_token(self) -> None:
        _user, social = self.do_refresh_token()
        self.assertEqual(social.extra_data["access_token"], "foobar-new-token")


class AzureADOAuth2TokenRequestBodyMixin(TestCase):
    def _token_request_body(self, url_prefix: str) -> dict[str, list[str]]:
        matches = [
            call.request
            for call in responses.calls
            if cast("str", call.request.url).startswith(url_prefix)
        ]
        self.assertGreaterEqual(
            len(matches),
            1,
            f"expected at least one token request for {url_prefix}, found {len(matches)}",
        )
        request = matches[-1]
        body = request.body or ""
        if isinstance(body, bytes):
            body = body.decode()
        return parse_qs(body)


class AzureADOAuth2FederatedIdentityCredentialTest(
    AzureADOAuth2TokenRequestBodyMixin, AzureADOAuth2Test
):
    def extra_settings(self):
        settings = super().extra_settings()
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_SECRET", None)
        settings["SOCIAL_AUTH_AZUREAD_OAUTH2_CLIENT_ASSERTION"] = "fic-assertion"
        return settings

    def test_login_uses_client_assertion(self) -> None:
        self.do_login()
        body = self._token_request_body(self.backend.access_token_url())
        self.assertIn("client_assertion", body)
        self.assertEqual(body["client_assertion"], ["fic-assertion"])
        self.assertEqual(
            body.get("client_assertion_type"),
            ["urn:ietf:params:oauth:client-assertion-type:jwt-bearer"],
        )
        self.assertNotIn("client_secret", body)

    def test_refresh_token_uses_client_assertion(self) -> None:
        self.do_refresh_token()
        body = self._token_request_body(self.backend.refresh_token_url())
        self.assertIn("client_assertion", body)
        self.assertNotIn("client_secret", body)


class AzureADOAuth2FederatedIdentityCredentialFromFileTest(
    AzureADOAuth2TokenRequestBodyMixin, AzureADOAuth2Test
):
    def setUp(self) -> None:
        super().setUp()
        # Default token file for class-level flows; individual tests can override.
        self.token_path = self._write_temp_token("default-assertion")
        patcher = patch.dict(
            os.environ,
            {
                "OAUTH2_FEDERATED_TOKEN_FILE": self.token_path,
                "AZURE_FEDERATED_TOKEN_FILE": self.token_path,
            },
            clear=False,
        )
        patcher.start()
        self.addCleanup(patcher.stop)

    def extra_settings(self):
        settings = super().extra_settings()
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_SECRET", None)
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_CLIENT_ASSERTION", None)
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_FEDERATED_TOKEN_FILE", None)
        return settings

    def _write_temp_token(self, value: str) -> str:
        with tempfile.NamedTemporaryFile("w", delete=False) as handle:
            handle.write(value)
            self.addCleanup(os.remove, handle.name)
            return handle.name

    def test_login_uses_oauth2_env_token_path(self) -> None:
        token_path = self._write_temp_token("env-assertion")
        with patch.dict(
            os.environ, {"OAUTH2_FEDERATED_TOKEN_FILE": token_path}, clear=False
        ):
            self.do_login()

        body = self._token_request_body(self.backend.access_token_url())
        self.assertEqual(body.get("client_assertion"), ["env-assertion"])
        self.assertEqual(
            body.get("client_assertion_type"),
            ["urn:ietf:params:oauth:client-assertion-type:jwt-bearer"],
        )
        self.assertNotIn("client_secret", body)

    def test_login_uses_azure_env_token_path(self) -> None:
        token_path = self._write_temp_token("azure-env-assertion")
        with patch.dict(
            os.environ,
            {
                "OAUTH2_FEDERATED_TOKEN_FILE": "",
                "AZURE_FEDERATED_TOKEN_FILE": token_path,
            },
            clear=False,
        ):
            self.do_login()

        body = self._token_request_body(self.backend.access_token_url())
        self.assertEqual(body.get("client_assertion"), ["azure-env-assertion"])
        self.assertNotIn("client_secret", body)

    def test_missing_env_token_path_raises(self) -> None:
        with (
            patch.dict(
                os.environ,
                {"OAUTH2_FEDERATED_TOKEN_FILE": "/no/such/file"},
                clear=False,
            ),
            self.assertRaises(AuthMissingParameter),
        ):
            self.do_login()

    def test_empty_token_file_raises(self) -> None:
        token_path = self._write_temp_token(" \n\t")
        with (
            patch.dict(
                os.environ,
                {
                    "OAUTH2_FEDERATED_TOKEN_FILE": token_path,
                    "AZURE_FEDERATED_TOKEN_FILE": token_path,
                },
                clear=False,
            ),
            self.assertRaises(AuthMissingParameter),
        ):
            self.do_login()

    def test_empty_token_file_optional_returns_none(self) -> None:
        token_path = self._write_temp_token(" \n\t")
        with patch.dict(
            os.environ,
            {
                "OAUTH2_FEDERATED_TOKEN_FILE": token_path,
                "AZURE_FEDERATED_TOKEN_FILE": token_path,
            },
            clear=False,
        ):
            assertion = self.backend.client_assertion(required=False)

        self.assertIsNone(assertion)


class AzureADOAuth2MissingCredentialsTest(AzureADOAuth2Test):
    def extra_settings(self):
        settings = super().extra_settings()
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_SECRET", None)
        settings.pop("SOCIAL_AUTH_AZUREAD_OAUTH2_CLIENT_ASSERTION", None)
        return settings

    def test_missing_secret_and_assertion_fails(self) -> None:
        with (
            patch.dict(
                os.environ,
                {
                    "AZURE_FEDERATED_TOKEN_FILE": "",
                    "OAUTH2_FEDERATED_TOKEN_FILE": "",
                },
                clear=False,
            ),
            self.assertRaises(AuthMissingParameter),
        ):
            self.do_login()

    def test_login(self) -> None:
        with self.assertRaises(AuthMissingParameter):
            super().test_login()

    def test_partial_pipeline(self) -> None:
        with self.assertRaises(AuthMissingParameter):
            super().test_partial_pipeline()

    def test_refresh_token(self) -> None:
        with self.assertRaises(AuthMissingParameter):
            super().test_refresh_token()
